-- This Source Code Form is subject to the terms of the bCDDL, v. 1.1.
-- If a copy of the bCDDL was not distributed with this
-- file, You can obtain one at http://beamng.com/bCDDL-1.1.txt

local M = {}
--Mandatory controller parameters
M.type = "main"
M.relevantDevice = nil
M.engineInfo = {}
M.gearboxInfo = {isReverse = false}
M.fireEngineTemperature = 0
M.throttle = 0
M.brake = 0
M.clutchRatio = 0
-----

local min = math.min
local max = math.max
local abs = math.abs
local floor = math.floor

local constants = {rpmToAV = 0.104719755, avToRPM = 9.549296596425384}

local leftEngine = nil
local rightEngine = nil
local leftPropShaft = nil
local rightPropShaft = nil

local rightWheel = nil
local leftWheel = nil

local rpmRange = 0
local idleRPM = 0

local targetThrustSetting = 0

local prevAlt = 0
local vertSpeedSmoothing = newExponentialSmoothing(30)
local bmepLSmoothing = newExponentialSmoothing(12)
local bmepRSmoothing = newExponentialSmoothing(12)

local leftIgnition = 0
local rightIgnition = 0
local blockIgniton = false

local energyStorageData = {
	tanks = {},
	ratio = 0,
	volume = 0,
	capacity = 0,
	invEnergyStorageCount = 0,
}

local function updateGFX(dt)
	targetThrustSetting = min(max(targetThrustSetting + (electrics.values.targetThrustSettingIncrease * dt * 0.5) - (electrics.values.targetThrustSettingDecrease * dt * 0.5), 0), 1)

	local brakeinput = electrics.values.brakeOverride or min(max(input.brake or 0, 0), 1)
	M.clutch = electrics.values.clutchOverride or min(max(input.clutch or 0, 0), 1)
	electrics.values.wheelspeed = electrics.values.airspeed

	M.throttle = targetThrustSetting
	
	local leftRpm = 0
	local rightRpm = 0

	local targetRpm = 2250

	if leftEngine then
		leftRpm = leftEngine.outputAV1 * constants.avToRPM or 0 
		electrics.values.leftClutch = (leftEngine.isStalled or leftEngine.ignitionCoef == 0) and 0 or 1
	end
	if rightEngine then
		rightRpm = rightEngine.outputAV1 * constants.avToRPM or 0 
		electrics.values.rightClutch = (rightEngine.isStalled or rightEngine.ignitionCoef == 0) and 0 or 1
	end

	local rpm = max(leftRpm, rightRpm)

	local rpmDiff = leftRpm - rightRpm
	
	local leftThrottle = 0
	local rightThrottle = 0

	if electrics.values.leftThrottleInput > M.throttle then leftThrottle = electrics.values.leftThrottleInput else leftThrottle = M.throttle end
	if electrics.values.rightThrottleInput > M.throttle then rightThrottle = electrics.values.rightThrottleInput else rightThrottle = M.throttle end

	M.throttle = max(leftThrottle, rightThrottle)

	M.fireEngineTemperature = max((leftEngine and leftEngine.thermals) and leftEngine.thermals.exhaustTemperature or 0, (rightEngine and rightEngine.thermals) and rightEngine.thermals.exhaustTemperature or 0) or obj:getEnvTemperature()

	energyStorageData.ratio = 0
	energyStorageData.volume = 0

	for _,s in pairs(energyStorageData.tanks) do
		local energyStorage = energyStorage.getStorage(s)
		if energyStorage then
			energyStorageData.ratio = energyStorageData.ratio + energyStorage.remainingRatio
			energyStorageData.volume = energyStorageData.volume + energyStorage.remainingVolume
			
			if energyStorage.name == "leftTank" then electrics.values.leftTankVolume = energyStorage.remainingVolume
			elseif energyStorage.name == "rightTank" then electrics.values.rightTankVolume = energyStorage.remainingVolume end
		end
	end
	energyStorageData.ratio = energyStorageData.ratio * energyStorageData.invEnergyStorageCount
	
	electrics.values.fuel = energyStorageData.ratio or 0
	electrics.values.lowfuel = electrics.values.fuel < 0.1
	electrics.values.fuelCapacity = energyStorageData.invEnergyStorageCount > 0 and (energyStorageData.capacity or 0) or 1
	electrics.values.fuelVolume = energyStorageData.volume

	if M.throttle == 0 and brakeinput == 0 then
		if electrics.values.airspeed > 15 then M.brake = 0.6 else M.brake = 0.12 end
	else M.brake = brakeinput end

	electrics.values.throttle = M.throttle
	electrics.values.brake = M.brake
	electrics.values.clutch = 0
	electrics.values.clutchRatio = 1
	electrics.values.isShifting = false
	electrics.values.gear = string.format("%d%%", M.throttle * 100)
	electrics.values.gear_A = 0
	electrics.values.gearModeIndex = 0
	electrics.values.gearIndex = 0
	electrics.values.idlerpm = idleRPM
	electrics.values.rpm = rpm
	electrics.values.rpmLeft = leftRpm
	electrics.values.rpmRight = rightRpm
	electrics.values.throttleLeft = leftThrottle
	electrics.values.throttleRight = rightThrottle
	electrics.values.wheelspeed = electrics.values.airspeed
	electrics.values.engineLoad = max((not leftEngine or leftEngine.isDisabled) and 0 or leftEngine.engineLoad, (not rightEngine or rightEngine.isDisabled) and 0 or rightEngine.engineLoad)
	
	electrics.values.oiltemp = max(leftEngine and leftEngine.thermals.oilTemperature or 0, rightEngine and rightEngine.thermals.oilTemperature or 0) or 0
	electrics.values.oiltempL = leftEngine and leftEngine.thermals.oilTemperature or 0
	electrics.values.oiltempR = rightEngine and rightEngine.thermals.oilTemperature or 0
	electrics.values.watertemp = electrics.values.oiltemp
	electrics.values.ignition = (leftEngine and leftEngine.ignitionCoef > 0 and not leftEngine.isDisabled) or (rightEngine and rightEngine.ignitionCoef > 0 and not rightEngine.isDisabled)
	electrics.values.running = (leftEngine and not leftEngine.isDisabled) or (rightEngine and not rightEngine.isDisabled) or 0
	electrics.values.engineRunning = electrics.values.running
	electrics.values.engineThrottle = max(leftThrottle, rightThrottle)
	electrics.values.checkengine = leftEngine and leftEngine.isDisabled or rightEngine and rightEngine.isDisabled or false
	electrics.values.engineLoad = max(leftEngine and leftEngine.instantEngineLoad or 0, rightEngine and rightEngine.instantEngineLoad or 0)

	electrics.values.propPitchLeft = max(min((leftRpm / targetRpm)^5 * 0.8, 1), 0)
	electrics.values.propPitchRight = max(min((rightRpm / targetRpm)^5 * 0.8, 1), 0)

	if (rightEngine and (rightEngine.isStalled or rightEngine.ignitionCoef == 0)) then
		electrics.values.propPitchRight = 1
	end

	if (leftEngine and (leftEngine.isStalled or leftEngine.ignitionCoef == 0)) then
		electrics.values.propPitchLeft = 1
	end

	if streams.willSend("engineInfo") then
		M.engineInfo[1] = max(leftEngine and leftEngine.idleRPM or 0, rightEngine and rightEngine.idleRPM or 0)
		M.engineInfo[2] = max(leftEngine and leftEngine.maxRPM or 0, rightEngine and rightEngine.idleRPM or 0)
		M.engineInfo[5] = rpm
		M.engineInfo[6] = string.format("%d%%", M.throttle * 100)
		M.engineInfo[7] = 0
		M.engineInfo[8] = 0
		M.engineInfo[9] = (leftEngine and leftEngine.outputTorque1 or 0) + (rightEngine and rightEngine.outputTorque1 or 0)
		M.engineInfo[10] = M.engineInfo[9] or 0
		M.engineInfo[11] = obj:getGroundSpeed() -- airspeed
		M.engineInfo[12] = electrics.values.fuelVolume * (electrics.values.ignitionLevel > 0 and 1 or 0)
		M.engineInfo[13] = electrics.values.fuelCapacity
		M.engineInfo[14] = "manual"
		M.engineInfo[16] = 0
		M.engineInfo[17] = 0
		M.engineInfo[18] = max(leftEngine and ((leftEngine.outputAV1 > leftEngine.starterMaxAV * 0.8) and 1) or 0, rightEngine and ((rightEngine.outputAV1 > rightEngine.starterMaxAV * 0.8) and 1 ) or 0) or 1
		M.engineInfo[19] = electrics.values.engineLoad
		M.engineInfo[20] = 0
		M.engineInfo[21] = 0
		M.engineInfo[22] = M.engineInfo[9] * rpm * constants.rpmToAV * 0.001 * 1.35962
	end

	-----instrument electrics-----
	local airspeed = electrics.values.airspeed
	if airspeed > 20.577 then
		electrics.values.airspeedAdjusted = 20.577 + ((airspeed - 20.577) * 4.15)
	else
		electrics.values.airspeedAdjusted = airspeed
	end

	local vertSpeed = (prevAlt - electrics.values.altitude) / -dt
	electrics.values.vertSpeed = vertSpeedSmoothing:get(vertSpeed * 196.85)
	if electrics.values.vertSpeed > 1000 then
		electrics.values.vertSpeed = 1000 + 0.5 * (electrics.values.vertSpeed - 1000)
	end
	prevAlt = electrics.values.altitude

	local dirVec = obj:getDirectionVector()
	electrics.values.pitch =  math.deg(math.atan(dirVec.z / math.sqrt(dirVec.x*dirVec.x + dirVec.y*dirVec.y)))

	local rollpitchyaw = {obj:getRollPitchYaw()}
	electrics.values.heading = rollpitchyaw[3] * 180 / math.pi + 180
	electrics.values.roll = math.asin(rollpitchyaw[1]) * 180 / math.pi

	electrics.values.bmepL = leftEngine and bmepLSmoothing:get(leftEngine.combustionTorque) or 0
	electrics.values.bmepR = rightEngine and bmepRSmoothing:get(rightEngine.combustionTorque) or 0
	
	-----Front Wheel steering-----
	electrics.values.wheelsteer = electrics.values.airspeed < 22 and electrics.values.rudder or 0
end

local function setStarter(enabled)
	if leftEngine and enabled and (leftEngine.outputAV1 < leftEngine.starterMaxAV * 0.75) then
		leftEngine:activateStarter()
	elseif leftEngine then
		leftEngine:deactivateStarter()
    end

	if rightEngine and enabled and (rightEngine.outputAV1 < rightEngine.starterMaxAV * 0.75) then
		rightEngine:activateStarter()
	elseif rightEngine then
		rightEngine:deactivateStarter()
    end
end

local function setOneStarter(enabled, side)
	if leftEngine and enabled and (leftEngine.outputAV1 < leftEngine.starterMaxAV * 0.75) and side == 0 and leftEngine.ignitionCoef == 1 then
		leftEngine:activateStarter()
	elseif leftEngine then
		leftEngine:deactivateStarter()
    end

	if rightEngine and enabled and (rightEngine.outputAV1 < rightEngine.starterMaxAV * 0.75) and side == 1 and rightEngine.ignitionCoef == 1 then
		rightEngine:activateStarter()
	elseif rightEngine then
		rightEngine:deactivateStarter()
    end
end

local function setEngineIgnition(value)
	if leftEngine and not blockIgniton then
		M.setOneEngineIgnition(value and 1 or 0, 0)
	end

	if rightEngine and not blockIgniton then
		M.setOneEngineIgnition(value and 1 or 0, 1)
	end
	blockIgniton = false
end

local function setOneEngineIgnition(value,side)
	if side == 0 then
		leftIgnition = value
		leftEngine:setIgnition(leftIgnition)
		print("Left ignition: "..leftIgnition)
	elseif side == 1 then
		rightIgnition = value
		rightEngine:setIgnition(rightIgnition)
		print("Right ignition: "..rightIgnition)
	end
end

local function setOneEngineIgnitionInput(value,side)
	if electrics.values.ignitionLevel < 1 then return end

	if leftEngine and side == 0 then
		setOneEngineIgnition(value, side)
		if leftIgnition == 0 and rightIgnition == 0 then
			electrics.setIgnitionLevel(1)
		end
	elseif rightEngine and side == 1 then
		setOneEngineIgnition(value, side)
		if leftIgnition == 0 and rightIgnition == 0 then
			electrics.setIgnitionLevel(1)
		end
	else return end
	
	if value == 1 then
		blockIgniton = true
		electrics.setIgnitionLevel(2)
	end
end

local function toggleOneEngineIgnition(side)
	local currentIgnitionLevel = false
	if electrics.values.ignitionLevel < 1 then return end

	if leftEngine and side == 0 then
		currentIgnitionLevel = leftIgnition
	elseif rightEngine and side == 1 then
		currentIgnitionLevel = rightIgnition
	else return end

	setOneEngineIgnition(-1*(currentIgnitionLevel-1),side)

	if currentIgnitionLevel == 0 then
		blockIgniton = true
		electrics.setIgnitionLevel(2)
	elseif leftIgnition == 0 and rightIgnition == 0 then
		electrics.setIgnitionLevel(1)
	end
end

local function init(jbeamData)
	blockIgniton = false
	leftIgnition = 0
	rightIgnition = 0
	
	input.state["aileron"] = { minLimit=-1, maxLimit=1, smootherKBD=newTemporalSmoothing(10, 10, nil, 0), smootherPAD=newTemporalSmoothing(10, 10, nil, 0)}
	input.state["elevator"] = { minLimit=-1, maxLimit=1, smootherKBD=newTemporalSmoothing(10, 10, nil, 0), smootherPAD=newTemporalSmoothing(10, 10, nil, 0)}
	input.state["rudder"] = { minLimit=-1, maxLimit=1, smootherKBD=newTemporalSmoothing(10, 10, nil, 0), smootherPAD=newTemporalSmoothing(10, 10, nil, 0)}

	vertSpeedSmoothing:reset()
	bmepLSmoothing:reset()

	electrics.values.targetThrustSettingIncrease = 0
	electrics.values.targetThrustSettingDecrease = 0
	electrics.values.leftThrottleInput = 0
	electrics.values.rightThrottleInput = 0
	electrics.values.ignitionLevel = 0
	electrics.values.airspeedAdjusted = 0
	electrics.values.heading = 0
	electrics.values.bmepL = 0
	electrics.values.bmepR = 0

	targetThrustSetting = 0

	leftEngine = powertrain.getDevice("leftEngine")
	rightEngine = powertrain.getDevice("rightEngine")
	leftPropShaft = powertrain.getDevice("propShaftLeft")
	rightPropShaft = powertrain.getDevice("propShaftRight")

	idleRPM = leftEngine and leftEngine.idleRPM or rightEngine and rightEngine.idleRPM or 0
	rpmRange = (leftEngine and leftEngine.maxRPM - idleRPM) or (rightEngine and rightEngine.maxRPM - idleRPM) or 0

	energyStorageData = {
		tanks = {},
		ratio = 0,
		volume = 0,
		capacity = 0,
		invEnergyStorageCount = 0,
	}

	if leftEngine then
		for _,s in pairs(leftEngine.energyStorage or {}) do
			table.insert(energyStorageData.tanks, s)
		end
	end
	
	if rightEngine then
		for _,s in pairs(rightEngine.energyStorage or {}) do
			table.insert(energyStorageData.tanks, s)
		end
	end

	local energyStorageCount = 0
	for _, s in pairs(energyStorageData.tanks or {}) do
		local energyStorage = energyStorage.getStorage(s)
		if energyStorage and energyStorage.type ~= "n2oTank" then
		energyStorageData.capacity = energyStorageData.capacity + energyStorage.capacity
		energyStorageCount = energyStorageCount + 1
		end
	end
	energyStorageData.invEnergyStorageCount = energyStorageCount > 0 and 1 / energyStorageCount or 0
	
	for _,wd in pairs(wheels.wheels) do
		if wd.name == "MR" then rightWheel = wd end
		if wd.name == "ML" then leftWheel = wd end
    end
	
	electrics.values.vertSpeed = 0
	prevAlt = 100
	
	--longer render distance
	obj:queueGameEngineLua(string.format("be:getObjectByID(%d):setField('renderDistance', '', '%d')", obj:getID(), 8000))
end

local function vehicleActivated()
	if not playerInfo.firstPlayerSeated then return end
	
	M.sendTorqueData()
end

local function sendTorqueData()
	if leftEngine then
		leftEngine:sendTorqueData()
	end
	if rightEngine then
		rightEngine:sendTorqueData()
	end
end

M.init = init
M.onReset = init
M.updateGFX = updateGFX

--Mandatory main controller API
M.shiftUp = nop
M.shiftDown = nop
M.shiftToGearIndex = nop
M.cycleGearboxModes = nop
M.setStarter = setStarter
M.freeze = nop
M.vehicleActivated = vehicleActivated
-------------------------------
M.sendTorqueData = sendTorqueData
M.setOneStarter = setOneStarter
M.setEngineIgnition = setEngineIgnition
M.setOneEngineIgnition = setOneEngineIgnition
M.setOneEngineIgnitionInput = setOneEngineIgnitionInput
M.toggleOneEngineIgnition = toggleOneEngineIgnition
return M
