local M = {}

local min = math.min
local max = math.max

local waterDoorsOpen = false
local leftWaterVolume = 0
local rightWaterVolume = 0
local waterVolume = 0
local isPickingUp = false
local pickUpRate = 0
local waterTimer = 0
local particleLifeTime = 4

local extParticles = {}
local potentialTargets = {}
local currentEmitter = 1

local function init(jbeamData)
	dropNodes = jbeamData.dropNodes_nodes
	pickUpNodes = jbeamData.pickUpNodes_nodes
	waterTankNodes = jbeamData.waterTankNodes_nodes
	overFlowNodes = jbeamData.overFlowNodes_nodes
	flowRate = jbeamData.flowRate
	maxVolume = jbeamData.maxVolume
	startVolume = jbeamData.startVolume
	pickUpRate = jbeamData.pickUpRate
	waterVolume = startVolume

	leftWaterVolume = startVolume / 2
	rightWaterVolume = startVolume / 2

	electrics.values.leftWaterVolume = leftWaterVolume
	electrics.values.rightWaterVolume = rightWaterVolume
	waterDoorsOpen = false
	
	obj:setNodeMass(waterTankNodes[1], 2+leftWaterVolume)
	obj:setNodeMass(waterTankNodes[2], 2+rightWaterVolume)
	
	waterTimer = 0
	
	extParticles = {}
end

--PRESS THE BOMB OVER--
local function waterRelease()
	if waterVolume > 0 and not isPickingUp then
		currentEmitter = 1
		waterDoorsOpen = true
		controller.getControllerSafe('waterDoorCoupler').detachGroup()
		
		local mapObjects = mapmgr.getObjects()
		--all vehicles outside a 200m radius will be ignored
		for id,v in pairs(mapObjects) do
		if id ~= obj:getId() and (vec3(obj:getPosition())-vec3(v.pos)):length() < 200 then
			potentialTargets[id] = v
		end
		end
	end
	--obj:queueGameEngineLua("local vid = be:getPlayerVehicleID(0) for i = be:getObjectCount()-1, 0, -1 do local veh = be:getObject(i) if veh:getId() ~= vid then veh:queueLuaCommand('fire.extinguishVehicle()') end end")
end

local function update(dt)
end
--for i,v in pairs(getmetatable(obj)[1]) do print(i) end
--obj:addParticle(vec3(989.329,-712.680,110.980),vec3(0,0,0),vec3(0,0,0),29,1,44)
local function updateGFX(dt)
	for i,v in pairs(potentialTargets) do
		potentialTargets[i] = mapmgr.getObjects()[i]
	end
	
	isPickingUp = false
	if (obj:inWater(pickUpNodes[1]) or obj:inWater(pickUpNodes[2])) and (obj:getAirSpeed() > 33) then
		isPickingUp = true
		if obj:inWater(pickUpNodes[1]) then
			leftWaterVolume = min(leftWaterVolume + pickUpRate*dt, maxVolume/2)
			if leftWaterVolume == maxVolume/2 then obj:addParticleByNodes(overFlowNodes[1], overFlowNodes[3], -1, 23, 1, 20) end
			end
		if obj:inWater(pickUpNodes[2]) then
			rightWaterVolume = min(rightWaterVolume + pickUpRate*dt, maxVolume/2) 
			if rightWaterVolume == maxVolume/2 then obj:addParticleByNodes(overFlowNodes[2], overFlowNodes[4], -1, 23, 1, 20) end
			end

		electrics.values.leftWaterVolume = leftWaterVolume
		electrics.values.rightWaterVolume = rightWaterVolume
		waterVolume = leftWaterVolume + rightWaterVolume
	end

	if waterDoorsOpen then
		if waterVolume <= 0 then waterTimer = 0 waterVolume = 0 waterDoorsOpen = false controller.getControllerSafe('waterDoorCoupler').toggleGroup()
		else
			M.emitTestParticles(dropNodes[1], dropNodes[2])
			leftWaterVolume = leftWaterVolume - flowRate * dt
			rightWaterVolume = rightWaterVolume - flowRate * dt
			waterTimer = waterTimer + dt
			if waterTimer >= 0.4 then waterTimer = 0 M.createParticle() end
		end

		if leftWaterVolume <= 0 then leftWaterVolume = 0 end
		if rightWaterVolume <= 0 then rightWaterVolume = 0 end

		electrics.values.leftWaterVolume = leftWaterVolume
		electrics.values.rightWaterVolume = rightWaterVolume
		waterVolume = leftWaterVolume + rightWaterVolume
	end

	local alreadyEmit = false
	for i,v in pairs(extParticles) do
		extParticles[i].lifeTime = extParticles[i].lifeTime-dt
		if extParticles[i].lifeTime <= 0 then extParticles[i] = nil else

		extParticles[i].vel = extParticles[i].vel + (obj:getGravityVector() * dt)
		extParticles[i].pos = extParticles[i].pos + extParticles[i].vel * dt
		if i > currentEmitter and not alreadyEmit then
			obj:addParticle(extParticles[i].pos,vec3(0,0,0),vec3(0,0,0),23,(particleLifeTime - extParticles[i].lifeTime)*3.5,46)
			if i == #extParticles then
				currentEmitter = 0
			else
			currentEmitter = i end
			alreadyEmit = true
		end
		for id,t in pairs(potentialTargets) do
			if (extParticles[i].pos-t.pos):length() < 12 then
				obj:queueGameEngineLua(string.format("for i = be:getObjectCount()-1, 0, -1 do local veh = be:getObject(i) if veh:getId()==%s then veh:queueLuaCommand('fire.extinguishVehicle()') end end",tostring(id)))
				local hit = true
			end
		end
		if hit then extParticles[i] = nil end
		end
	end

	obj:setNodeMass(waterTankNodes[1], 2+leftWaterVolume)
	obj:setNodeMass(waterTankNodes[2], 2+rightWaterVolume)
end

local function createParticle()
	extParticles[#extParticles+1] = {
		pos = obj:getNodePosition(dropNodes[2]) + obj:getPosition(),
		vel = obj:getVelocity(),
		lifeTime = particleLifeTime
	}
end

local function emitTestParticles(node1, node2)
	obj:addParticleByNodes(node1, node2, 21, 13, 1, 6)
	obj:addParticleByNodes(node1, node2, 2, 22, 1, 6)
end

M.init = init
M.waterRelease = waterRelease
M.emitTestParticles = emitTestParticles
M.createParticle = createParticle
M.updateGFX = updateGFX
M.update = update
return M